# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import uuid

import packaging.version

from pifpaf import drivers


class CephDriver(drivers.Driver):
    DEFAULT_PORT = 6790

    def __init__(self, port=DEFAULT_PORT,
                 **kwargs):
        """Create a new Ceph cluster."""
        super(CephDriver, self).__init__(**kwargs)
        self.port = port

    @classmethod
    def get_options(cls):
        return [
            {"param_decls": ["--port"],
             "type": int,
             "default": cls.DEFAULT_PORT,
             "help": "port to use for Ceph Monitor"},
        ]

    def _setUp(self):
        super(CephDriver, self)._setUp()

        self._ensure_xattr_support()

        fsid = str(uuid.uuid4())
        conffile = os.path.join(self.tempdir, "ceph.conf")
        mondir = os.path.join(self.tempdir, "mon", "ceph-a")
        osddir = os.path.join(self.tempdir, "osd", "ceph-0")
        os.makedirs(mondir)
        os.makedirs(osddir)

        _, version = self._exec(["ceph", "--version"], stdout=True)
        version = version.decode("ascii").split()[2]
        version = packaging.version.Version(version)

        if version < packaging.version.Version("12.0.0"):
            extra = """
mon_osd_nearfull_ratio = 1
mon_osd_full_ratio = 1
osd_failsafe_nearfull_ratio = 1
osd_failsafe_full_ratio = 1
"""
        else:
            extra = """
mon_allow_pool_delete = true
"""

        # Enable msgrv2 protocol if Nautilus or later.
        if version >= packaging.version.Version("14.0.0"):
            msgrv2_extra = """
mon_host = v2:127.0.0.1:%(port)d/0
""" % dict(port=self.port)
        else:
            msgrv2_extra = ""

        # FIXME(sileht): check availible space on /dev/shm
        # if os.path.exists("/dev/shm") and os.access('/dev/shm', os.W_OK):
        #     journal_path = "/dev/shm/$cluster-$id-journal"
        # else:
        journal_path = "%s/osd/$cluster-$id/journal" % self.tempdir

        with open(conffile, "w") as f:
            f.write("""[global]
fsid = %(fsid)s
%(msgrv2_extra)s

# no auth for now
auth cluster required = none
auth service required = none
auth client required = none

## no replica
osd pool default size = 1
osd pool default min size = 1
osd crush chooseleaf type = 0

## some default path change
run dir = %(tempdir)s
pid file = %(tempdir)s/$type.$id.pid
max open files = 1024
admin socket = %(tempdir)s/$cluster-$name.asok
mon data = %(tempdir)s/mon/$cluster-$id
osd data = %(tempdir)s/osd/$cluster-$id
osd journal = %(journal_path)s
log file = %(tempdir)s/$cluster-$name.log
mon cluster log file = %(tempdir)s/$cluster.log

# Only omap to have same behavior for all filesystems
filestore xattr use omap = True

# workaround for ext4 and last Jewel version
osd max object name len = 256
osd max object namespace len = 64
osd op threads = 10
filestore max sync interval = 10001
filestore min sync interval = 10000

# disable POOL_NO_REDUNDANCY warning
mon_warn_on_pool_no_redundancy = false

%(extra)s

journal_aio = false
journal_dio = false
journal zero on create = false
journal block align = false

# run as file owner
setuser match path = %(tempdir)s/$type/$cluster-$id

[mon]
mon pg warn min per osd = 0
mon data avail warn = 2
mon data avail crit = 1

# Ceph CVE CVE-2021-20288 to prevent HEALTH_WARN
auth allow insecure global id reclaim = false

[mon.a]
host = localhost
mon addr = 127.0.0.1:%(port)d
""" % dict(fsid=fsid, msgrv2_extra=msgrv2_extra, tempdir=self.tempdir,
           port=self.port, journal_path=journal_path, extra=extra))  # noqa

        ceph_opts = ["ceph", "-c", conffile]
        mon_opts = ["ceph-mon", "-c", conffile, "--id", "a", "-d"]
        mgr_opts = ["ceph-mgr", "-c", conffile, "-d"]
        osd_opts = ["ceph-osd", "-c", conffile, "--id", "0", "-d",
                    "-m", "127.0.0.1:%d" % self.port]

        # Create and start monitor
        self._exec(mon_opts + ["--mkfs"])
        self._touch(os.path.join(mondir, "done"))
        mon, _ = self._exec(
            mon_opts,
            wait_for_line=r"mon.a@0\(leader\).mds e1 print_map")

        # Start manager
        if version >= packaging.version.Version("12.0.0"):
            self._exec(
                mgr_opts,
                wait_for_line="(mgr send_beacon active|waiting for OSDs)")

        # Create and start OSD
        self._exec(ceph_opts + ["osd", "create"])
        self._exec(ceph_opts + ["osd", "crush", "add", "osd.0", "1",
                                "root=default"])
        self._exec(osd_opts + ["--mkfs", "--mkjournal"])
        if version < packaging.version.Version("0.94.0"):
            wait_for_line = "journal close"
        else:
            wait_for_line = "done with init"
        osd, _ = self._exec(osd_opts, wait_for_line=wait_for_line)

        if version >= packaging.version.Version("12.0.0"):
            self._exec(ceph_opts + ["osd", "set-full-ratio", "0.95"])
            self._exec(ceph_opts + ["osd", "set-backfillfull-ratio", "0.95"])
            self._exec(ceph_opts + ["osd", "set-nearfull-ratio", "0.95"])

        # Wait it's ready
        out = b""
        while b"HEALTH_OK" not in out:
            ceph, out = self._exec(ceph_opts + ["health"], stdout=True)
            if b"HEALTH_ERR" in out:
                raise RuntimeError("Fail to deploy ceph")

        self.putenv("CEPH_CONF", conffile, True)
        self.putenv("CEPH_CONF", conffile)
        self.putenv("URL", "ceph://localhost:%d" % self.port)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import shutil
import urllib.parse
import uuid

import click

from pifpaf import drivers
from pifpaf.drivers import postgresql
from pifpaf.drivers import redis

LOG = logging.getLogger(__name__)


class GnocchiDriver(drivers.Driver):

    DEFAULT_PORT = 8041
    DEFAULT_PORT_INDEXER = 9541
    DEFAULT_PORT_COORDINATOR = 9542

    def __init__(self, port=DEFAULT_PORT, indexer_port=DEFAULT_PORT_INDEXER,
                 statsd_port=None,
                 indexer_url=None,
                 storage_url=None,
                 coordination_driver="default",
                 coordination_port=DEFAULT_PORT_COORDINATOR,
                 **kwargs):
        """Create a new Gnocchi instance."""
        super(GnocchiDriver, self).__init__(**kwargs)
        self.port = port
        self.indexer_port = indexer_port
        self.indexer_url = indexer_url
        self.storage_url = storage_url
        self.statsd_port = statsd_port
        self.coordination_driver = coordination_driver
        self.coordination_port = coordination_port

    @classmethod
    def get_options(cls):
        return [
            {"param_decls": ["--port"],
             "type": int,
             "default": cls.DEFAULT_PORT,
             "help": "port to use for Gnocchi HTTP API"},
            {"param_decls": ["--statsd-port"],
             "type": int,
             "help": "port to use for gnocchi-statsd"},
            {"param_decls": ["--indexer-port"],
             "type": int,
             "default": cls.DEFAULT_PORT_INDEXER,
             "help": "port to use for Gnocchi indexer"},
            {"param_decls": ["--coordination-port"],
             "type": int,
             "default": cls.DEFAULT_PORT_COORDINATOR,
             "help": "port to use for Gnocchi coordination"},
            {"param_decls": ["--coordination-driver"],
             "type": click.Choice(["default", "redis"]),
             "default": "default",
             "help": "Select a coordination driver"},
            {"param_decls": ["--indexer-url"],
             "help": "indexer URL to use"},
            {"param_decls": ["--storage-url"],
             "help": "storage URL to use"},
        ]

    def _setUp(self):
        super(GnocchiDriver, self)._setUp()

        try:
            shutil.copy(self.find_config_file("gnocchi/api-paste.ini"),
                        self.tempdir)
        except RuntimeError:
            pass
        try:
            shutil.copy(self.find_config_file("gnocchi/policy.json"),
                        self.tempdir)
        except RuntimeError:
            pass

        if self.indexer_url is None:
            pg = self.useFixture(
                postgresql.PostgreSQLDriver(port=self.indexer_port))
            self.indexer_url = pg.url

        if self.storage_url is None:
            self.storage_url = "file://%s" % self.tempdir

        conffile = os.path.join(self.tempdir, "gnocchi.conf")

        storage_parsed = urllib.parse.urlparse(self.storage_url)
        storage_driver = storage_parsed.scheme

        if storage_driver == "s3":
            storage_config = {
                "s3_access_key_id": (urllib.parse.unquote(
                    storage_parsed.username or "S3RVER")),
                "s3_secret_access_key": (
                    urllib.parse.unquote(
                        storage_parsed.password or "S3RVER")),
                "s3_endpoint_url": "http://%s:%s/%s" % (
                    storage_parsed.hostname,
                    storage_parsed.port,
                    storage_parsed.path,
                )
            }
        elif storage_driver == "swift":
            storage_config = {
                "swift_authurl": "http://%s:%s%s" % (
                    storage_parsed.hostname,
                    storage_parsed.port,
                    storage_parsed.path,
                ),
                "swift_user": (urllib.parse.unquote(
                    storage_parsed.username or "admin:admin")),
                "swift_key": (urllib.parse.unquote(
                    storage_parsed.password or "admin")),
            }
        elif storage_driver == "ceph":
            storage_config = {
                "ceph_conffile": storage_parsed.path,
            }
        elif storage_driver == "redis":
            storage_config = {
                "redis_url": self.storage_url,
            }
        elif storage_driver == "file":
            storage_config = {
                "file_basepath": (
                    storage_parsed.path or self.tempdir
                ),
            }
        else:
            raise RuntimeError("Storage driver %s is not supported" %
                               storage_driver)

        if self.coordination_driver == "redis":
            r = self.useFixture(redis.RedisDriver(port=self.coordination_port))
            storage_config["coordination_url"] = r.url

        storage_config_string = "\n".join(
            "%s = %s" % (k, v)
            for k, v in storage_config.items()
        )
        statsd_resource_id = str(uuid.uuid4())

        with open(conffile, "w") as f:
            f.write("""[DEFAULT]
debug = %s
verbose = True
[api]
host = localhost
port = %s
[storage]
driver = %s
%s
[metricd]
metric_processing_delay = 1
metric_cleanup_delay = 1
[statsd]
resource_id = %s
creator = admin
user_id = admin
project_id = admin
[indexer]
url = %s""" % (self.debug,
               self.port,
               storage_driver,
               storage_config_string,
               statsd_resource_id,
               self.indexer_url))

        self._exec(["gnocchi-upgrade", "--config-file=%s" % conffile])

        c, _ = self._exec(["gnocchi-metricd", "--config-file=%s" % conffile],
                          wait_for_line="metrics wait to be processed")

        c, _ = self._exec(["gnocchi-statsd", "--config-file=%s" % conffile],
                          wait_for_line=("(Resource .* already exists"
                                         "|Created resource )"))

        args = ["gnocchi-api", "--config-file=%s" % conffile]
        c, _ = self._exec(args,
                          wait_for_line=r"WSGI app 0 \(mountpoint=''\) ready")
        self.addCleanup(self._kill, c)

        self.http_url = "http://localhost:%d" % self.port

        self.putenv("GNOCCHI_PORT", str(self.port))
        self.putenv("URL", "gnocchi://localhost:%d" % self.port)
        self.putenv("GNOCCHI_HTTP_URL", self.http_url)
        self.putenv("GNOCCHI_ENDPOINT", self.http_url, True)
        self.putenv("OS_AUTH_TYPE", "gnocchi-basic", True)
        self.putenv("GNOCCHI_STATSD_RESOURCE_ID", statsd_resource_id, True)
        self.putenv("GNOCCHI_USER", "admin", True)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import logging
import os

import psutil

LOG = logging.getLogger(__name__)


def _get_procs_of_pgid(wanted_pgid):
    procs = []
    for p in psutil.process_iter():
        try:
            pgid = os.getpgid(p.pid)
        except OSError as e:
            # ESRCH is returned if process just died in the meantime
            if e.errno != errno.ESRCH:
                raise
            continue
        if pgid == wanted_pgid:
            procs.append(p)
    return procs


def process_cleaner(parent):
    do_sigkill = False
    # NOTE(sileht): Add processes from process tree and process group
    # Relying on process tree only will not work in case of
    # parent dying prematuraly and double fork
    # Relying on process group only will not work in case of
    # subprocess calling again setsid()
    procs = set(_get_procs_of_pgid(parent.pid))
    try:
        LOG.debug("Terminating %s (%s)",
                  " ".join(parent.cmdline()), parent.pid)
        procs |= set(parent.children(recursive=True))
        procs.add(parent)
        parent.terminate()
    except psutil.NoSuchProcess:
        LOG.warning("`%s` is already gone, sending SIGKILL to its process "
                    "group", parent)
        do_sigkill = True
    else:
        # Waiting for all processes to stop
        for p in procs:
            try:
                LOG.debug("Waiting %s (%s)", " ".join(p.cmdline()), p.pid)
            except psutil.NoSuchProcess:
                pass
        gone, alive = psutil.wait_procs(procs, timeout=10)
        if alive:
            do_sigkill = True
            LOG.warning("`%s` didn't terminate cleanly after 10 seconds, "
                        "sending SIGKILL to its process group", parent)

    if do_sigkill and procs:
        for p in procs:
            try:
                LOG.debug("Killing %s (%s)", " ".join(p.cmdline()), p.pid)
                p.kill()
            except psutil.NoSuchProcess:
                pass
        gone, alive = psutil.wait_procs(procs, timeout=10)
        if alive:
            LOG.warning("`%s` survive SIGKILL", alive)

    # Be sure to get returncode
    parent.wait()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import os
import signal

from pifpaf import drivers


class RabbitMQDriver(drivers.Driver):

    DEFAULT_PORT = 5682
    DEFAULT_NODENAME = "pifpaf"
    DEFAULT_USERNAME = "pifpaf"
    DEFAULT_PASSWORD = "secret"

    def __init__(self, port=DEFAULT_PORT, nodename=DEFAULT_NODENAME,
                 username=DEFAULT_USERNAME, password=DEFAULT_PASSWORD,
                 cluster=False,
                 **kwargs):
        """Create a new RabbitMQ server."""
        super(RabbitMQDriver, self).__init__(**kwargs)
        self.port = port
        self.nodename = nodename
        self.username = username
        self.password = password
        self.cluster = cluster
        self._path = ["/usr/lib/rabbitmq/bin/",
                      "/usr/local/sbin"]
        self._process = {}
        self._ports = {}
        self._next_port = itertools.count(self.port)

    @classmethod
    def get_options(cls):
        return [
            {"param_decls": ["--port"],
             "type": int,
             "default": cls.DEFAULT_PORT,
             "help": "port to use for RabbitMQ"},
            {"param_decls": ["--nodename"],
             "default": cls.DEFAULT_NODENAME,
             "help": "RabbitMQ node name"},
            {"param_decls": ["--cluster"],
             "is_flag": True,
             "help": "Create a 3 HA node clusters"},
            {"param_decls": ["--username"],
             "default": cls.DEFAULT_USERNAME,
             "help": "RabbitMQ username"},
            {"param_decls": ["--password"],
             "default": cls.DEFAULT_PASSWORD,
             "help": "RabbitMQ password"},
        ]

    def get_port(self, nodename):
        if nodename not in self._ports:
            self._ports[nodename] = next(self._next_port)
        return self._ports[nodename]

    def start_node(self, nodename):
        port = self.get_port(nodename)
        if nodename in self._process:
            self.rabbitmqctl(nodename, ["start_app"])
        else:
            complete_env = {
                "RABBITMQ_NODE_IP_ADDRESS": b"127.0.0.1",
                "RABBITMQ_NODE_PORT": str(port),
                "RABBITMQ_NODENAME": nodename,
                "RABBITMQ_PIDFILE": os.path.join(self.tempdir,
                                                 nodename, "pid"),
            }
            complete_env.update(self.env)
            c, _ = self._exec(["rabbitmq-server"], env=complete_env,
                              path=self._path,
                              wait_for_line=("completed with .* plugins"))
            self.addCleanup(self.kill_node, nodename, ignore_not_exists=True)
            self._process[nodename] = c
        return port

    def stop_node(self, nodename):
        if nodename not in self._process:
            raise RuntimeError("node %s not started" % nodename)
        self.rabbitmqctl(nodename, ["stop_app"])

    def kill_node(self, nodename, signal=signal.SIGTERM,
                  ignore_not_exists=False):
        if nodename not in self._process:
            if not ignore_not_exists:
                raise RuntimeError("node %s not started" % nodename)
            return

        c = self._process.pop(nodename)
        try:
            os.killpg(c.pid, signal)
            os.waitpid(c.pid, 0)
        except OSError:
            pass

    def join_cluster(self, nodename, master):
        if master not in self._process:
            raise RuntimeError("node %s not started" % master)
        self.stop_node(nodename)
        self.rabbitmqctl(nodename, ["join_cluster", master])
        self.start_node(nodename)

    def rabbitmqctl(self, nodename, command):
        self._exec(["rabbitmqctl", "-n", nodename] + command,
                   path=self._path, env=self.env)

    def _setUp(self):
        super(RabbitMQDriver, self)._setUp()
        self.env = {
            "RABBITMQ_ENABLED_PLUGINS_FILE": os.path.join(self.tempdir,
                                                          "notexists"),
            "RABBITMQ_LOG_BASE": self.tempdir,
            "RABBITMQ_MNESIA_BASE": self.tempdir,
            "HOME": self.tempdir,
        }

        if self.cluster:
            n1 = self.nodename + "-1@localhost"
            n2 = self.nodename + "-2@localhost"
            n3 = self.nodename + "-3@localhost"
            # Start master
            self.start_node(n1)
            self.start_node(n2)
            self.start_node(n3)
            self.join_cluster(n2, n1)
            self.join_cluster(n3, n1)
        else:
            n1 = self.nodename + "@localhost"
            self.start_node(n1)

        self.rabbitmqctl(n1, ["add_user", self.username, self.password])
        self.rabbitmqctl(n1, ["set_permissions", self.username,
                              ".*", ".*", ".*"])

        self.putenv("RABBITMQ_HOME", self.tempdir)
        self.putenv("RABBITMQ_PORT", str(self.port))
        if self.cluster:
            self.putenv("RABBITMQ_NODENAME", n1)
            self.putenv("RABBITMQ_NODENAME1", n1)
            self.putenv("RABBITMQ_NODENAME2", n2)
            self.putenv("RABBITMQ_NODENAME3", n3)
            self.putenv("URL", "rabbit://"
                        "%(username)s:%(password)s@localhost:%(port1)d,"
                        "%(username)s:%(password)s@localhost:%(port2)d,"
                        "%(username)s:%(password)s@localhost:%(port3)d//" % {
                            'username': self.username,
                            'password': self.password,
                            'port1': self.port,
                            'port2': self.port + 1,
                            'port3': self.port + 2})
        else:
            self.putenv("RABBITMQ_NODENAME", n1)
            self.putenv("URL", "rabbit://%s:%s@localhost:%d//" % (
                self.username, self.password, self.port))
